package com.oracle.truffle.r.nodes.qirinterface;

import com.oracle.truffle.api.source.Source;
import com.oracle.truffle.api.source.SourceSection;
import com.oracle.truffle.r.nodes.RSyntaxNodeVisitor;
import com.oracle.truffle.r.nodes.access.AccessArgumentNode;
import com.oracle.truffle.r.nodes.access.ConstantNode;
import com.oracle.truffle.r.nodes.access.WriteLocalFrameVariableNode;
import com.oracle.truffle.r.nodes.access.variables.ReadVariableNode;
import com.oracle.truffle.r.nodes.control.*;
import com.oracle.truffle.r.nodes.function.FunctionExpressionNode;
import com.oracle.truffle.r.nodes.function.RCallNode;
import com.oracle.truffle.r.nodes.function.RCallSpecialNode;
import com.oracle.truffle.r.nodes.query.*;
import com.oracle.truffle.r.runtime.RRuntime;
import com.oracle.truffle.r.runtime.context.RContext;
import com.oracle.truffle.r.runtime.nodes.RNode;
import com.oracle.truffle.r.runtime.nodes.RSyntaxNode;

import qir.ast.*;
import qir.ast.operator.*;

/**
 * This visitor translates a {@link RNode} into a {@link QIRNode}.
 */
public final class QIRTranslateVisitor implements RSyntaxNodeVisitor<QIRNode> {
    private final RContext context;

    public QIRTranslateVisitor(RContext context) {
        this.context = context;
    }

    private final SourceSection dummy = Source.newBuilder("").name("QIR translate").mimeType(RRuntime.R_APP_MIME).build().createUnavailableSection();

    @Override
    public final QIRNode visit(final IfNode ifNode) {
        return new QIRIf(ifNode.getSourceSection(), ifNode.getCondition().asRSyntaxNode().accept(this), ifNode.getThenPart().asRSyntaxNode().accept(this),
                        ifNode.getElsePart().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final WhileNode whileNode) {
        return new QIRTruffleNode(whileNode.getSourceSection(), whileNode.getSourceSection().getCode());
    }

    @Override
    public final QIRNode visit(final ForNode forNode) {
        return new QIRTruffleNode(forNode.getSourceSection(), forNode.getSourceSection().getCode());
    }

    /**
     * Translation of a sequence of statements. Note: This is also where the STRAD-ASSIGN rules are
     * handled.
     */
    @Override
    public final QIRNode visit(final BlockNode block) {
        QIRNode result = null;
        final RNode[] children = block.getSequence();
        final int len = children.length;

        if (len == 0) // Block is empty
            return QIRNull.getInstance();

        result = children[len - 1].asRSyntaxNode().accept(this); // The last statement of the block
        /*
         * In this loop, we wrap the result with the other statements of the block in reverse order
         * in the way described by the STRAD-ASSIGN rules.
         */
        for (int i = len - 2; i > -1; i--) {
            final RNode child = children[i];
            if (child instanceof WriteLocalFrameVariableNode) { // STRAD-ASSIGN-ID
                final Object name = ((WriteLocalFrameVariableNode) child).getName();
                final RNode value = ((WriteLocalFrameVariableNode) child).getRhs();

                // If the assignment value reads an argument, then we translate to a lambda.
                result = new QIRLambda(dummy, null, new QIRVariable(dummy, (String) name, null), result);
                // Else we apply STRAD-ASSIGN-ID normally
                if (!(value instanceof AccessArgumentNode))
                    result = new QIRApply(dummy, result, value.asRSyntaxNode().accept(this));
            } else
                throw new UnsupportedOperationException("Cannot translate \"" + block.getSourceSection() + "\" to QIR: untranslatable sequence.");
        }
        return result;
    }

    @Override
    public final QIRNode visit(final ReadVariableNode var) {
        return new QIRVariable(var.getSourceSection(), var.getIdentifier(), null);
    }

    @Override
    public final QIRNode visit(final WriteLocalFrameVariableNode var) {
        // STRAD-ASSIGN-ID should be handled in BlockNode.
        throw new RuntimeException("Error in translation to QIR: should not have visited an assignment.");
    }

    @Override
    public final QIRNode visit(final FunctionExpressionNode fun) {
        throw new RuntimeException("Can't translate R function to QIR.");
    }

    @Override
    public final QIRNode visit(final RCallNode call) {
        final QIRNode fun = call.getFunctionNode().asRSyntaxNode().accept(this);
        final RSyntaxNode[] args = call.getArguments().getArguments();
        final int nbArgs = args.length;
        QIRNode res = fun;

        if (nbArgs == 0) // call is an application with no arguments
            return new QIRApply(dummy, fun, null);
        for (final RSyntaxNode arg : args)
            res = new QIRApply(dummy, res, arg.accept(this));
        return res;
    }

    @Override
    public final QIRNode visit(final RCallSpecialNode call) {
        throw new RuntimeException("Can't translate R special call to QIR.");
    }

    @Override
    public final QIRNode visit(final ReplacementDispatchNode repl) {
        throw new RuntimeException("Can't translate R replacement node to QIR.");
    }

    /**
     * Translation of a subquery.
     */
    @Override
    public final QIRNode visit(final RQIRWrapperNode query) {
        return context.queries.get(query.id);
    }

    @Override
    public final QIRNode visit(final RSelectNode select) {
        return new QIRProject(select.getSourceSection(), select.formatter.asRSyntaxNode().accept(this), select.query.asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final RFromNode from) {
        return new QIRScan(from.getSourceSection(), from.getTable().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final RWhereNode where) {
        return new QIRFilter(where.getSourceSection(), where.getFilter().asRSyntaxNode().accept(this), where.getQuery().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final RGroupNode group) {
        return new QIRGroupBy(group.getSourceSection(), group.getGroup().asRSyntaxNode().accept(this), group.getQuery().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final ROrderNode order) {
        return new QIRSortBy(order.getSourceSection(), order.getOrder().asRSyntaxNode().accept(this), order.getIsAscending().asRSyntaxNode().accept(this),
                        order.getQuery().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final RJoinNode join) {
        return new QIRJoin(join.getSourceSection(), join.getFilter().asRSyntaxNode().accept(this), join.getLeft().asRSyntaxNode().accept(this), join.getRight().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final RLeftJoinNode join) {
        return new QIRLeftJoin(join.getSourceSection(), join.getFilter().asRSyntaxNode().accept(this), join.getLeft().asRSyntaxNode().accept(this), join.getRight().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final RRightJoinNode join) {
        return new QIRRightJoin(join.getSourceSection(), join.getFilter().asRSyntaxNode().accept(this), join.getLeft().asRSyntaxNode().accept(this), join.getRight().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final RLimitNode limit) {
        return new QIRLimit(limit.getSourceSection(), limit.getLimit().asRSyntaxNode().accept(this), limit.getQuery().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final ConstantNode cst) {
        return QIRInterface.RToQIRType(cst.getSourceSection(), cst.getValue());
    }
}